<?php

namespace Civi\StripeImport;

use Civi\Payment\Exception\PaymentProcessorException;

abstract class Base {

  /**
   * @var int The payment processor ID
   */
  private int $paymentProcessorID;

  /**
   * @var \CRM_Core_Payment_Stripe|\CRM_Core_Payment
   */
  protected $paymentProcessor;

  public function setPaymentProcessorID(int $paymentProcessorID) {
    $this->paymentProcessorID = $paymentProcessorID;
    $this->paymentProcessor = \Civi\Payment\System::singleton()->getById($this->getPaymentProcessorID());
    if (!isset($this->paymentProcessor->stripeClient)) {
      throw new PaymentProcessorException('stripeClient is not initialised - maybe payment processor is not configured correctly?');
    }
  }

  public function getPaymentProcessorID(): int {
    return $this->paymentProcessorID;
  }

  public function getPaymentProcessorObject(): \CRM_Core_Payment {
    return $this->paymentProcessor;
  }

  /**
   * @var int|NULL The CiviCRM Contact ID
   */
  protected int $contactID;

  public function setContactID($contactID) {
    $this->contactID = $contactID;
  }

  public function getContactID(): int {
    return $this->contactID;
  }

  public function hasContactID(): bool {
    return isset($this->contactID);
  }

  protected string $description = '';

  public function setDescription(string $description) {
    $this->description = $description;
  }

  public function getDescription(): string {
    return $this->description;
  }

  protected int $paymentInstrumentID;

  public function setPaymentInstrumentID(int $paymentInstrumentID) {
    $this->paymentInstrumentID = $paymentInstrumentID;
  }

  public function getPaymentInstrumentID(): int {
    return $this->paymentInstrumentID;
  }

  protected int $financialTypeID;

  public function setFinancialTypeID(int $financialTypeID) {
    $this->financialTypeID = $financialTypeID;
  }

  public function getFinancialTypeID(): int {
    return $this->financialTypeID;
  }

  /**
   *
   * @param \Stripe\StripeObject|PropertySpy $stripeObject
   *
   * @return void
   * @throws \CRM_Core_Exception
   * @throws \Civi\API\Exception\UnauthorizedException
   * @throws \Stripe\Exception\ApiErrorException
   */
  protected function deriveContactIDFromCustomer($stripeObject, bool $createMissingContact = TRUE): void {
    if (!$this->hasContactID()) {
      // Derive contact ID from stripe customer
      if (empty(\CRM_Stripe_Api::getObjectParam('customer_id', $stripeObject))) {
        throw new \Exception('Missing customer_id in Stripe Object. Cannot derive contact ID');
      }

      $importer = new \Civi\StripeImport\Customer();
      $importer->setPaymentProcessorID($this->getPaymentProcessorID());
      $importer->setStripeCustomerID(\CRM_Stripe_Api::getObjectParam('customer_id', $stripeObject));
      $importer->setCreateMissingContact($createMissingContact);
      $return = $importer->importOne();
      $this->setContactID($return['contact_id']);
    }
  }

  /**
   * Disable the "updateMembershipBasedOnCompletionOfContribution()"
   * @return void
   */
  protected function disableUpdateMembershipBasedOnCompletionOfContribution() {
    if (\Civi\Core\Container::singleton()->has('civi_membership_order_complete')) {
      \Civi::service('civi_membership_order_complete')->setActive(FALSE);
      \Civi::log()->info('Stripeimport: Disabled Civi\Membership\OrderCompleteSubscriber::onOrderComplete listener');
    }
    else {
      \Civi::log()->debug('civi_membership_order_complete service not found. Did you apply https://github.com/civicrm/civicrm-core/pull/30085?');
    }
  }

  /**
   * Import all Stripe Objects
   * This function calls importMultiple() multiple times using the 'continue_after'
   *   API filter to iterate through all objects
   *
   * @return array
   * @throws \Exception
   */
  public function importAll(): array {
    $stripeAPIFilters = [
      'limit' => 100,
    ];
    $results = $this->importMultiple($stripeAPIFilters);

    while ($results['continue_after']) {
      $starting_after = $results['continue_after'];

      $stripeAPIFilters = [
        'limit' => 100,
        'starting_after' => $starting_after,
      ];
      $additional = $this->importMultiple($stripeAPIFilters);

      $results['imported'] = array_merge($results['imported'], $additional['imported']);
      $results['skipped'] = array_merge($results['skipped'], $additional['skipped']);
      $results['errors'] = array_merge($results['errors'], $additional['errors']);
      $results['count'] = [
        'imported' => count($results['imported']),
        'skipped' => count($results['skipped']),
        'errors' => count($results['errors']),
      ];

      if ($additional['continue_after']) {
        $results['continue_after'] = $additional['continue_after'];
      } else {
        unset($results['continue_after']);
      }
    }
    unset($results['continue_after']);
    return $results;
  }

}
