<?php

namespace Civi\StripeImport;

use Civi\Api4\Contribution;
use Civi\StripeImport\Event\StartImportEvent;
use CRM_StripeImport_ExtensionUtil as E;

class Charge extends Base {

  // @todo: handle charge.refunded?

  /**
   * @var string The Stripe charge ID
   */
  private string $stripeChargeID;

  public function setStripeChargeID(string $ChargeID) {
    $this->stripeChargeID = $ChargeID;
  }

  public function getStripeChargeID(): string {
    return $this->stripeChargeID;
  }

  /**
   * @var int The CiviCRM Contribution ID
   */
  private int $contributionID;

  public function setContributionID(int $contributionID) {
    $this->contributionID = $contributionID;
  }

  public function getContributionID(): int {
    return $this->contributionID;
  }

  public function hasContributionID(): bool {
    return isset($this->contributionID);
  }

  /**
   * If running importOne multiple times we have to reset the class vars
   *
   * @return void
   */
  public function resetOne() {
    unset($this->contactID);
    unset($this->contributionID);
    $this->setDescription('');
  }

  /**
   * Import a single charge
   *
   * @return array
   * @throws \CRM_Core_Exception
   * @throws \Civi\API\Exception\UnauthorizedException
   * @throws \Civi\Payment\Exception\PaymentProcessorException
   * @throws \Stripe\Exception\ApiErrorException
   */
  public function importOne(): array {
    \Civi::log('stripeimport')->info('StripeImport: ' . $this->getStripeChargeID() . ': START');
    \Civi::dispatcher()->dispatch(
      'civi.stripeimport.startimport',
      new StartImportEvent(
        $this->getPaymentProcessorID(),
        $this->getStripeChargeID(),
        'charge',
      )
    );

    // Get the Charge from Stripe.
    try {
      $stripeCharge = $this->paymentProcessor->stripeClient->charges->retrieve($this->getStripeChargeID());
    }
    catch (\Exception $e) {
      throw new \CRM_Core_Exception('Error retrieving existing charge: ' . $e->getMessage());
    }

    // Get the related invoice (to see if we have a subscription)
    $stripeInvoiceID = \CRM_Stripe_Api::getObjectParam('invoice_id', $stripeCharge);
    if (!empty($stripeInvoiceID)) {
      try {
        $stripeInvoice = $this->paymentProcessor->stripeClient->invoices->retrieve($stripeCharge->invoice);
      }
      catch (\Exception $e) {
        throw new \CRM_Core_Exception('Error retrieving existing invoice for charge: ' . $e->getMessage());
      }
    }

    // Check for a subscription.
    if (isset($stripeInvoice)) {
      $subscriptionID = \CRM_Stripe_Api::getObjectParam('subscription_id', $stripeInvoice);
      if (!empty($subscriptionID)) {
        if (\Civi::settings()->get('stripeimport_chargeimport_subscriptioncharges')) {
          \Civi::log('stripeimport')->info('StripeImport: ' . $this->getStripeChargeID() . ': Importing charge linked to subscription: ' . $subscriptionID);
        }
        else {
          \Civi::log('stripeimport')->info('StripeImport: ' . $this->getStripeChargeID() . ': Ignoring. It is linked to subscription: ' . $subscriptionID);
          $results = [
            'charge' => $this->getStripeChargeID(),
            'subscription' => $subscriptionID,
            'status' => 'ignored',
          ];
          \Civi::log('stripeimport')->info('StripeImport: ' . $this->getStripeChargeID() . ': END ' . print_r($results, TRUE));
          return $results;
        }
      }
    }

    // Make sure we have a Contact ID (if necessary one will be created)
    $this->deriveContactIDFromCustomer($stripeCharge);

    // Get the description for the charge
    $this->getDescription();
    if (empty($this->getDescription())) {
      $this->setDescription(\CRM_Stripe_Api::getObjectParam('description', $stripeCharge) ?? '');
    }
    if (empty($this->getDescription())) {
      $this->setDescription('StripeImport Charge');
    }
    \Civi::log('stripeimport')->info('StripeImport: ' . $stripeCharge->id . ': Description: "' . $this->getDescription() . '"');

    // Prepare to either create or update a contribution record in CiviCRM.
    $contributionParams = [];

    // We update these parameters regardless if it's a new contribution
    // or an existing contributions.
    $contributionParams['receive_date'] = \CRM_Stripe_Api::getObjectParam('receive_date', $stripeCharge);
    $contributionParams['total_amount'] = $this->paymentProcessor->api->getValueFromStripeObject('amount', 'Float', $stripeCharge);

    // Check if a contribution already exists.
    if (!$this->hasContributionID()) {
      // See if we have a matching payment for this trxn_id and get the linked contribution ID
      $contribution = civicrm_api3('Mjwpayment', 'get_contribution', [
        'trxn_id' => $this->getStripeChargeID(),
      ]);
      if (!empty($contribution['id'])) {
        $this->setContributionID($contribution['id']);
      }
    }

    // If no contribution was found we need to create it
    if (!$this->hasContributionID()) {
      // We have to build all the parameters.
      $contributionParams['contact_id'] = $this->getContactID();
      $contributionParams['currency'] = $this->paymentProcessor->api->getValueFromStripeObject('currency', 'String', $stripeCharge);
      $contributionParams['trxn_id'] = $this->getStripeChargeID();
      $contributionParams['payment_instrument_id'] = $this->getPaymentInstrumentID();
      $contributionParams['financial_type_id'] = $this->getFinancialTypeID();
      $contributionParams['is_test'] = $this->paymentProcessor->getPaymentProcessor()['is_test'];
      $contributionParams['contribution_source'] = $this->getDescription();
      $contributionParams['contribution_status_id:name'] = 'Pending';
      if (!empty($contributionRecurID)) {
        $contributionParams['contribution_recur_id'] = $contributionRecurID;
      }
      $contribution = Contribution::create(FALSE)
        ->setValues($contributionParams)
        ->execute()
        ->first();
      $this->setContributionID($contribution['id']);
    }

    $webhookEventProcessor = new \Civi\Stripe\Webhook\Events($this->getPaymentProcessorID());
    $webhookEventProcessor->setData(\Stripe\StripeObject::constructFrom(['object' => $stripeCharge]));
    switch (\CRM_Stripe_Api::getObjectParam('status', $stripeCharge)) {
      case 'succeeded':
        \Civi::log('stripeimport')->info('StripeImport: contributionID: ' . $this->getContributionID() . ' Triggering charge.succeeded webhook event');
        $webhookEventProcessor->setEventType('charge.succeeded');
        $webhookEventProcessor->setSendEmailReceipt(\Civi::settings()->get('stripeimport_chargeimport_sendreceipt'));
        $webhookResult = $webhookEventProcessor->doChargeSucceeded();
        \Civi::log('stripeimport')->info('StripeImport: contributionID: ' . $this->getContributionID() . ' charge.succeeded result: ' . ($webhookResult->ok ? 'SUCCESS' : 'ERROR') . ' Message: ' . $webhookResult->message);
        break;

      case 'failed':
        \Civi::log('stripeimport')->info('StripeImport: contributionID: ' . $this->getContributionID() . ' Triggering charge.failed webhook event');
        $webhookEventProcessor->setEventType('charge.failed');
        $webhookResult = $webhookEventProcessor->doChargeFailed();
        \Civi::log('stripeimport')->info('StripeImport: contributionID: ' . $this->getContributionID() . ' charge.failed result: ' . ($webhookResult->ok ? 'SUCCESS' : 'ERROR') . ' Message: ' . $webhookResult->message);
        break;

      case 'pending':
        // Nothing further to do (we already created the contribution and should not create a payment or failure yet)
        $webhookResult = $webhookEventProcessor->getResultObject();
        $webhookResult->ok = TRUE;
        $webhookResult->message = 'Charge is Pending';
    }

    $results = [
      'charge' => $this->getStripeChargeID(),
      'customer' => \CRM_Stripe_Api::getObjectParam('customer_id', $stripeCharge),
      'contact_id' => $this->getContactID(),
      'contribution_id' => $this->getContributionID(),
      // 'status' => (isset($webhookResult) ? ($webhookResult->ok ? 'success' : 'error')), //@todo? What wrong
      'status_message' => $webhookResult->message ?? '',
    ];
    \Civi::log('stripeimport')->info('StripeImport: ' . $this->getStripeChargeID() . ': END ' . print_r($results, TRUE));
    return $results;
  }

  /**
   * @param array $stripeAPIFilters
   *   eg. ['limit' => 100, 'created' => ['>' => 20231202000000], 'starting_after' => 'sub_xxx']
   *   'created' accepts anything that strtotime accepts - eg. '-1 week'
   *
   * @return array
   * @throws \Exception
   */
  public function importMultiple(array $stripeAPIFilters = []): array {
    // Get the charges from Stripe
    if (empty($stripeAPIFilters['limit'])) {
      $stripeAPIFilters['limit'] = 100;
    }

    // @todo: CHECK the filter params for stripe charges
    foreach ($stripeAPIFilters as $filterKey => $filterValue) {
      switch ($filterKey) {
        case 'created_after':
          if (is_array($filterValue)) {
            throw new \CRM_Core_Exception('created_after cannot be an array');
          }
          // Don't add a blank filter value
          if (empty($filterValue)) {
            unset($stripeAPIFilters['created_after']);
            break;
          }

          $stripeAPIFilters['created']['gt'] = strtotime($filterValue);
          unset($stripeAPIFilters['created_after']);
          // $stripeAPIFilters['created']['gte'];
          break;

        case 'created_before':
          if (is_array($filterValue)) {
            throw new \CRM_Core_Exception('created_after cannot be an array');
          }
          // Don't add a blank filter value
          if (empty($filterValue)) {
            unset($stripeAPIFilters['created_before']);
            break;
          }

          $stripeAPIFilters['created']['lt'] = strtotime($filterValue);
          // $stripeAPIFilters['created']['lte'];
          unset($stripeAPIFilters['created_before']);
          break;

        default:
          if (empty($filterValue)) {
            unset($stripeAPIFilters[$filterKey]);
          }
      }

    }

    $stripeCharges = $this->getPaymentProcessorObject()->stripeClient->charges->all($stripeAPIFilters);

    $stripeChargeIDs = array_map(
      function ($charge) { return $charge->id; },
      $stripeCharges->data
    );

    // Prepare to collect the results
    $results = [
      'imported' => [],
      'errors' => [],
      'continue_after' => NULL
    ];

    // Exit if there aren't records to process
    if (!count($stripeChargeIDs)) {
      return $results;
    }

    foreach ($stripeChargeIDs as $chargeID) {
      $this->resetOne();
      $this->setStripeChargeID($chargeID);

      try {
        $resultImportOne = $this->importOne();
        $results['imported'][$chargeID] = $resultImportOne;
      }
      catch (\Exception $e) {
        \Civi::log('stripeimport')->error('StripeImport: ' . $chargeID . ' Import Failed: ' . $e->getMessage());
        $results['errors'][$chargeID] = $e->getMessage();
      }
    }
    $results['continue_after'] = $chargeID;
    $results['count'] = [
      'imported' => count($results['imported']),
      'errors' => count($results['errors']),
    ];
    return $results;
  }

  /**
   * @param \Stripe\StripeObject|PropertySpy $stripeObject
   * @param array $result
   *
   * @return array
   * @throws \CRM_Core_Exception
   * @throws \Civi\API\Exception\UnauthorizedException
   * @throws \Civi\Payment\Exception\PaymentProcessorException
   * @throws \Stripe\Exception\ApiErrorException
   */
  public function handleWebhookEventNotMatched($stripeObject, array $result): array {
    \Civi::log('stripeimport')->info('StripeImport Charge::handleWebhookEventNotMatched: contribution_not_found');

    $createContact = $findContribution = $createContribution = FALSE;
    if (\Civi::settings()->get('stripeimport_notmatched_customer_create')) {
      $createContact = TRUE;
    }
    if (\Civi::settings()->get('stripeimport_notmatched_charge_find')) {
      $findContribution = TRUE;
    }
    if (\Civi::settings()->get('stripeimport_notmatched_charge_create')) {
      $createContribution = TRUE;
    }

    \Civi::log('stripeimport')->debug('StripeImport: Config: ' . ', createContact: ' . $createContact . ', findContribution: ' . $findContribution . ', createContribution: ' . $createContribution);

    if ($findContribution || $createContribution) {
      // Make sure we have a Contact ID
      $this->deriveContactIDFromCustomer($stripeObject, $createContact);
      if (!$this->hasContactID()) {
        \Civi::log('stripeimport')->error('StripeImport: Could not find contactID and create contact is disabled');
        return $result;
      }
    }

    if ($findContribution) {
      // Try to match an existing Contribution in CiviCRM
      // API needs them in YmdHis format. They come in as timestamp.
      $searchStartDate = date('Ymd', \CRM_Stripe_Api::getObjectParam('receive_date', $stripeObject)) . '000000';
      $searchEndDate = date('Ymd', \CRM_Stripe_Api::getObjectParam('receive_date', $stripeObject)) . '235959';

      // Look for a contribution:
      //   - That is "Pending"
      //   - That is NOT a template
      //   - That has a receive_date within the same "day" (24 hours from 0000 to 2359)
      //   - trxn_id has not been set
      //   - That has the most recent `receive_date`. If we already completed the next contribution we won't match..
      //       ..possibly we don't need this but it adds extra "safety" to the matching by reducing the scope for error.
      $contribution = Contribution::get(FALSE)
        ->addWhere('is_template', '=', FALSE)
        ->addWhere('contribution_status_id:name', '=', 'Pending')
        ->addWhere('receive_date', 'BETWEEN', [$searchStartDate, $searchEndDate])
        ->addWhere('trxn_id', 'IS NULL')
        ->addWhere('total_amount', '=', $this->paymentProcessor->api->getValueFromStripeObject('amount', 'Float', $stripeObject))
        ->addOrderBy('receive_date', 'DESC')
        ->execute()
        ->first();
      if (!empty($contribution)) {
        // We found a contribution so assign it back to the result so it can be used by the calling code.
        $contributionID = $contribution['id'];
        $trxnID = $this->getStripeChargeID() ?? NULL;
        if (!empty($trxnID)) {
          // Update the found contribution with the invoice/charge ID.
          $contribution = Contribution::update(FALSE)
            ->addWhere('id', '=', $contributionID)
            ->addValue('trxn_id', $trxnID);
          $contribution->execute()->first();
          \Civi::log('stripeimport')->info("StripeImport: Added missing trxnID {$trxnID} to contribution {$contributionID}");
        }
        $this->setContributionID($contribution['id']);
      }
    }

    if ($createContribution || ($this->hasContributionID())) {
      // Running the importer will create a contribution if it does not exist.
      // So we check if we either have "createContribution" enable or we found a contribution.
      $result = $this->importOne();
      \Civi::log('stripeimport')->info('StripeImport: Charge import ran with result: ' . json_encode($result));
      // @todo: Get the contribution as an array (API4 result)?
      $result['contribution'] = $contribution;
    }
    return $result;
  }

}
