<?php

namespace Civi\StripeImport;

use Civi\Api4\Contact;
use Civi\Api4\Email;
use Civi\Api4\StripeCustomer;

class Customer extends Base {

  /**
   * @var string The Stripe customer ID
   */
  private string $stripeCustomerID;

  public function setStripeCustomerID(string $customerID) {
    $this->stripeCustomerID = $customerID;
  }

  public function getStripeCustomerID(): string {
    return $this->stripeCustomerID;
  }

  private bool $createMissingContact;

  public function setCreateMissingContact(bool $createMissingContact) {
    $this->createMissingContact = $createMissingContact;
  }

  public function isCreateMissingContact(): bool {
    return $this->createMissingContact ?? TRUE;
  }

  public function importOne(): array {
    // Get the payment processor and activate the Stripe API
    $customer = $this->paymentProcessor->stripeClient->customers->retrieve($this->getStripeCustomerID());

    $return = [
      'stripe_id' => $customer->id,
      'name' => $customer->name ?? NULL,
      'email' => $customer->email ?? NULL,
    ];
    $civicrmStripeCustomers = StripeCustomer::get(FALSE)
      ->addWhere('customer_id', '=', $customer->id)
      ->execute();
    if ($civicrmStripeCustomers->count() > 0) {
      $return['message'] = 'customer already exists';
      $return['skipped'] = 1;
      return array_merge($return, $civicrmStripeCustomers->first());
    }

    // No existing Stripe Customer in CiviCRM
    $contact_id = NULL;

    // Search contact by email
    if ($customer->email || $customer->name) {
      // Stripe has a single field "name" but we need to split into first_name/last_name for CiviCRM.
      // This regex will split the name into first and last names based on the first space found.
      $re = '/^([^\s]*)\s(.*)$/m';
      preg_match_all($re, $customer->name, $matches, PREG_SET_ORDER, 0);
      if (empty($matches)) {
        // No space so we'll use the full name as "first name".
        $firstName = $customer->name;
        $lastName = '';
      }
      else {
        $firstName = isset($matches[0][1]) ? $matches[0][1] : '';
        $lastName = isset($matches[0][2]) ? $matches[0][2] : '';
      }

      if(!$customer->email) {
        // No point in searching for an existing contact.
        $contactIDs = [];
      }
      else {
        // This will return any matching email in the system
        $civicrmEmails = Email::get(FALSE)
          ->addWhere('email', '=', $customer->email)
          ->addWhere('contact_id.is_deleted', '=', FALSE)
          ->execute();

        // List of contact IDs using this email address
        $contactIDsForEmail = array_map(
          function ($found_email) { return $found_email['contact_id']; },
          $civicrmEmails->getArrayCopy()
        );
        $contactIDs = $contactIDsForEmail;
      }

      if (empty($contactIDs) && $this->isCreateMissingContact()) {
        // Create the new contact record
        $newContact = Contact::create(FALSE)
          ->addValue('contact_type', 'Individual')
          ->addValue('source', 'Stripe > ' . $customer->description)
          ->addValue('first_name', $firstName)
          ->addValue('last_name', $lastName)
          ->execute()->first();

        if ($customer->email) {
          Email::create(FALSE)
            ->addValue('contact_id', $newContact['id'])
            ->addValue('email', $customer->email)
            ->execute();
        }
        $contact_id = $newContact['id'];
        $return['message'] = 'contact created';
      }
      elseif (count($contactIDs) === 1) {
        $contact_id = end($contactIDs);
        // Report the contact as found by email
        $return['skipped'] = 1;
        \Civi::log('stripeimport')->debug('stripeimport: Contact: ' . $contact_id . ' matched to customer: ' . $customer->id . ' by email.');
        $return['message'] = 'contact matched to email';
      }
      else if (count($contactIDs) > 1) {
        $contact_id = end($contactIDs);
        // Report the contact as duplicated
        $return['dupes'] = $contactIDs;
        \Civi::log('stripeimport')->debug('stripeimport: Multiple contacts found for customer: ' . $customer->id . '. Mapping to contact ID: ' . $contact_id);
        $return['message'] = 'contact matched to multiple contacts';
      }
      else {
        \Civi::log('stripeimport')->debug('stripeimport: No contact IDs and create contact is disabled for customer: ' . $customer->id);
        $return['message'] = 'no contact';
      }
    }
    $return['contact_id'] = $contact_id;

    // Try to create the Stripe customer record
    try {
      $newStripeCustomer = StripeCustomer::create(FALSE)
        ->addValue('customer_id', $customer->id)
        ->addValue('processor_id', $this->getPaymentProcessorID())
        ->addValue('currency', $customer->currency);
      if (!empty($contact_id)) {
        $newStripeCustomer->addValue('contact_id', $contact_id);
      }
      return array_merge($return, $newStripeCustomer->execute()->first());
    } catch (\Exception $e) {
      // Keep running if it already existed
      \Civi::log('stripeimport')->error('Could not create StripeCustomer in CiviCRM: ' . $e->getMessage());
    }
    return $return;
  }

  public function importMultiple(int $limit = 100, array $stripeAPIFilters = []): array {
    // Prepare an array to collect the results
    $results = [
      'imported' => [],
      'skipped' => [],
      'errors' => [],
      'continue_after' => NULL,
    ];

    // Get customers from Stripe
    if (empty($stripeAPIFilters['limit'])) {
      $stripeAPIFilters['limit'] = $limit;
    }
    foreach ($stripeAPIFilters as $filterKey => $filterValue) {
      switch ($filterKey) {
        case 'starting_after':
          if (empty($filterValue)) {
            unset($stripeAPIFilters[$filterValue]);
          }
          break;

        case 'created':
          // Don't add a blank filter value
          if (empty($filterValue)) {
            unset($stripeAPIFilters[$filterValue]);
            break;
          }
          // Check if we have an array with single key (we support 4 operators)
          if (!is_array($filterValue) || count($filterValue) !== 1) {
            throw new \Exception('stripeimport: "created" parameter is malformed');
          }
          switch (array_key_first($filterValue)) {
            case '>':
              $stripeAPIFilters['created']['gt'] = strtotime(reset($filterValue));
              break;

            case '>=':
              $stripeAPIFilters['created']['gte'] = strtotime(reset($filterValue));
              break;

            case '<':
              $stripeAPIFilters['created']['lt'] = strtotime(reset($filterValue));
              break;

            case '<=':
              $stripeAPIFilters['created']['lte'] = strtotime(reset($filterValue));
              break;

            default:
              throw new \Exception('stripeimport: "created" unsupported operator: ' . array_key_first($filterValue));
          }
      }

    }

    $customers_stripe = $this->paymentProcessor->stripeClient->customers->all($stripeAPIFilters);

    // Exit if there aren't records to process
    if (count($customers_stripe->data) === 0) {
      return $results;
    }

    $results['customers_count'] = count($customers_stripe->data);

    // Search the customers in CiviCRM
    $customer_ids = array_map(
      function ($customer) {
        return $customer->id;
      },
      $customers_stripe->data
    );
    $customers_stripe_clean = $customers_stripe->data;

    if (count($customer_ids)) {
      $customers_in_civicrm = StripeCustomer::get(FALSE)
        ->addWhere('customer_id', 'IN', $customer_ids)
        ->addWhere('contact_id', 'IS NOT NULL')
        ->execute()
        ->indexBy('customer_id');
    }
    else {
      $customers_in_civicrm = [];
    }

    foreach ($customers_stripe_clean as $customer) {
      $results['continue_after'] = $customer->id;

      // Return if contact was found
      if (isset($customers_in_civicrm[$customer->id])) {
        $results['skipped'][] = [
          'contact_id' => $customers_in_civicrm[$customer->id]['contact_id'],
          'email' => $customer->email,
          'stripe_id' => $customer->id,
        ];
        continue;
      }

      $this->setStripeCustomerID($customer->id);
      $importedCustomer = $this->importOne();
      $data = [
        'contact_id' => $importedCustomer['contact_id'],
        'email' => $importedCustomer['email'],
        'stripe_id' => $importedCustomer['stripe_id'],
        'message' => $importedCustomer['message'],
      ];
      if (array_key_exists('skipped', $importedCustomer)) {
        $index = 'skipped';
      }
      elseif (array_key_exists('dupes', $importedCustomer)) {
        $index = 'errors';
        $data['message'] = $importedCustomer['message'];
      }
      else {
        $index = 'imported';
      }
      $results[$index][] = $data;
    }
    return $results;
  }

  public function importAll(): array {
    $limit = 100;
    $imported = $this->importMultiple($limit);

    while ($imported['continue_after']) {
      $starting_after = $imported['continue_after'];

      $stripeAPIFilters = [
        'starting_after' => $starting_after,
      ];
      $additional = $this->importMultiple($limit, $stripeAPIFilters);

      $imported['imported'] = array_merge($imported['imported'], $additional['imported']);
      $imported['skipped'] = array_merge($imported['skipped'], $additional['skipped']);
      $imported['errors'] = array_merge($imported['errors'], $additional['errors']);
      $imported['customers_count'] += $additional['customers_count'];

      if ($additional['continue_after']) {
        $imported['continue_after'] = $additional['continue_after'];
      } else {
        unset($imported['continue_after']);
      }
    }
    unset($imported['continue_after']);
    return $imported;
  }

}
