<?php

namespace Civi\StripeImport;

use Brick\Money\Money;
use Civi\Api4\Contribution;
use Civi\Api4\ContributionRecur;
use Civi\Api4\Membership;
use Civi\StripeImport\Event\MembershipAutoEvent;
use Civi\StripeImport\Event\StartImportEvent;
use CRM_StripeImport_ExtensionUtil as E;

class Subscription extends Base {

  /**
   * @var string The Stripe subscription ID
   */
  private string $stripeSubscriptionID;

  public function setStripeSubscriptionID(string $subscriptionID) {
    $this->stripeSubscriptionID = $subscriptionID;
  }

  public function getStripeSubscriptionID(): string {
    return $this->stripeSubscriptionID;
  }

  /**
   * @var int The CiviCRM ContributionRecur ID
   */
  private int $contributionRecurID;

  public function setContributionRecurID(int $contributionRecurID) {
    $this->contributionRecurID = $contributionRecurID;
  }

  public function getContributionRecurID(): int {
    return $this->contributionRecurID;
  }

  public function hasContributionRecurID(): bool {
    return isset($this->contributionRecurID);
  }

  private bool $emailReceipt = FALSE;

  public function setEmailReceipt(bool $emailReceipt) {
    $this->emailReceipt = $emailReceipt;
  }

  public function getEmailReceipt(): bool {
    return $this->emailReceipt;
  }

  /**
   * @var int The CiviCRM Membership ID
   */
  private int $membershipID;

  public function setMembershipID(int $membershipID) {
    $this->membershipID = $membershipID;
  }

  public function getMembershipID(): int {
    return $this->membershipID;
  }

  public function hasMembershipID(): bool {
    return isset($this->membershipID);
  }

  private bool $membershipAuto = FALSE;

  public function setMembershipAuto(bool $membershipAuto) {
    $this->membershipAuto = $membershipAuto;
  }

  public function isMembershipAuto(): bool {
    return $this->membershipAuto;
  }

  private bool $membershipDisableUpdates = FALSE;

  public function setMembershipDisableUpdates(bool $membershipDisableUpdates) {
    $this->membershipDisableUpdates = $membershipDisableUpdates;
  }

  public function isMembershipDisableUpdates(): bool {
    return $this->membershipDisableUpdates;
  }

  /**
   * Optional parameter to exclude certain Stripe subscription statuses from import
   * See https://docs.stripe.com/api/subscriptions/object#subscription_object-status
   *
   * @var array
   */
  private array $subscriptionExcludeStatuses = [];

  public function getSubscriptionExcludeStatuses(): array {
    return $this->subscriptionExcludeStatuses;
  }

  /**
   * @param array|string $subscriptionExcludeStatuses
   *
   * @return void
   */
  public function setSubscriptionExcludeStatuses($subscriptionExcludeStatuses) {
    if (!is_array($subscriptionExcludeStatuses)) {
      $subscriptionExcludeStatuses = explode(',', $subscriptionExcludeStatuses);
    }
    $this->subscriptionExcludeStatuses = $subscriptionExcludeStatuses;
  }

  /**
   * If running importOne multiple times we have to reset the class vars
   *
   * @return void
   */
  public function resetOne() {
    unset($this->contactID);
    unset($this->membershipID);
    unset($this->contributionRecurID);
    $this->setDescription('');
  }

  /**
   * Import a single subscription
   *
   * @return array
   * @throws \CRM_Core_Exception
   * @throws \Civi\API\Exception\UnauthorizedException
   * @throws \Civi\Payment\Exception\PaymentProcessorException
   * @throws \Stripe\Exception\ApiErrorException
   */
  public function importOne(): array {
    \Civi::log('stripeimport')->info('StripeImport: ' . $this->getStripeSubscriptionID() . ': START');
    \Civi::dispatcher()->dispatch(
      'civi.stripeimport.startimport',
      new StartImportEvent(
        $this->getPaymentProcessorID(),
        $this->getStripeSubscriptionID(),
        'subscription',
      )
    );

    // Get the subscription from Stripe
    try {
      $stripeSubscription = $this->getPaymentProcessorObject()->stripeClient->subscriptions->retrieve(
        $this->getStripeSubscriptionID(),
        ['expand' => ['items.data.plan.product']]
      );
    }
    catch (\Exception $e) {
      throw new \CRM_Core_Exception('Error retrieving existing subscription: ' . $e->getMessage());
    }

    // Check subscription status
    if (in_array($stripeSubscription->status, $this->getSubscriptionExcludeStatuses())) {
      $results = [
        'subscription' => $this->getStripeSubscriptionID(),
        'customer' => \CRM_Stripe_Api::getObjectParam('customer_id', $stripeSubscription),
      ];
      \Civi::log('stripeimport')->info('StripeImport: ' . $this->getStripeSubscriptionID() . ': Not importing because subscription status "' . $stripeSubscription->status . '" is excluded.');
      \Civi::log('stripeimport')->info('StripeImport: ' . $this->getStripeSubscriptionID() . ': END ' . print_r($results, TRUE));
      return $results;
    }

    // Make sure we have a Contact ID (if necessary one will be created)
    $this->deriveContactIDFromCustomer($stripeSubscription);

    // Need currency, frequency_unit, frequency_interval for checking currency/amount which is calculated from items on Stripe subscription
    // Need contribution_status_id:name for checking recur status matches what is at Stripe
    $contributionRecurApi = ContributionRecur::get(FALSE)
      ->addSelect('currency', 'frequency_unit', 'frequency_interval', 'contribution_status_id:name')
      ->addWhere('is_test', '=', $this->getPaymentProcessorObject()->getPaymentProcessor()['is_test']);

    if (!$this->hasContributionRecurID()) {
      // See if we can find an existing recur by Stripe Subscription ID
      $contributionRecurApi->addWhere('processor_id', '=', $this->getStripeSubscriptionID());
    }
    else {
      // Match on existing Recur ID
      $contributionRecurApi->addWhere('id', '=', $this->getContributionRecurID());
    }

    $contributionRecur = $contributionRecurApi->execute()->first();
    if (!empty($contributionRecur)) {
      // Great! We've found an existing contributionRecur that we can use.
      $this->setContributionRecurID($contributionRecur['id']);
    }

    if ($this->hasContributionRecurID()) {
      // Normally the amount update on the recur will be handled automatically by events/initial creation.
      // But if there is a mismatch we'll identify and fix here by creating a new template contribution and updating the recur amount.

      ///////////////////////////////////////////
      // Check: Status and update if necessary //
      ///////////////////////////////////////////
      $stripeRecurStatus = \CRM_Stripe_Api::mapSubscriptionStatusToRecurStatus($stripeSubscription->status);
      if ($contributionRecur['contribution_status_id:name'] !== $stripeRecurStatus) {
        \Civi::log('stripeimport')->warning('StripeImport: ' . $this->getStripeSubscriptionID() . ' recurID: ' . $this->getContributionRecurID()
          . ': Recur Status "' . $contributionRecur['contribution_status_id:name'] . '" does not match subscription status: ' . \CRM_Stripe_Api::mapSubscriptionStatusToRecurStatus($stripeSubscription->status));

        switch ($stripeRecurStatus) {
          case 'Cancelled':
            $cancelParams = [
              'id' => $contributionRecur['id'],
              'cancel_date' => \CRM_Stripe_Api::getObjectParam('cancel_date', $stripeSubscription),
            ];
            civicrm_api3('ContributionRecur', 'cancel', $cancelParams);
            break;

          default:
            ContributionRecur::update(FALSE)
              ->addWhere('id', '=', $contributionRecur['id'])
              ->addValue('contribution_status_id:name', $stripeRecurStatus)
              ->execute();
        }
        \Civi::log('stripeimport')
          ->info('StripeImport: ' . $this->getStripeSubscriptionID() . ' recurID: ' . $this->getContributionRecurID() . ': new status: ' . $stripeRecurStatus);
      }

      ///////////////////////////////////////////
      // Check: Amount and update if necessary //
      ///////////////////////////////////////////
      $calculatedItems = $this->getPaymentProcessorObject()->api->calculateItemsForSubscription($this->getStripeSubscriptionID(), $stripeSubscription->items->data);

      // Now check if recurring contribution matches frequency
      $contributionRecurKey = mb_strtolower($contributionRecur['currency']) . "_{$contributionRecur['frequency_unit']}_{$contributionRecur['frequency_interval']}";
      if (isset($calculatedItems[$contributionRecurKey])) {
        $calculatedItem = $calculatedItems[$contributionRecurKey];
        $templateContribution = \CRM_Contribute_BAO_ContributionRecur::getTemplateContribution($contributionRecur['id']);
        // We shouldn't get here without a template contribution existing. But if we do the amount comparison will fail
        // It seems this can happen sometimes when performing a new import.
        if (empty($templateContribution)) {
          \Civi::log('stripeimport')->error('StripeImport: ' . $this->getStripeSubscriptionID() . ' recurID: ' . $this->getContributionRecurID() . ': Missing template contribution! Skipping amount changed check');
        }
        elseif (!Money::of($calculatedItem['amount'], mb_strtoupper($calculatedItem['currency']))
          ->isAmountAndCurrencyEqualTo(Money::of($templateContribution['total_amount'], $templateContribution['currency']))) {
          // Create a new template contribution to update the amount
          ContributionRecur::updateAmountOnRecurMJW(FALSE)
            ->addWhere('id', '=', $contributionRecur['id'])
            ->addValue('amount', $calculatedItem['amount'])
            ->execute();
          \Civi::log('stripeimport')
            ->info('StripeImport: ' . $this->getStripeSubscriptionID() . ' recurID: ' . $this->getContributionRecurID() . ': new amount: ' . $calculatedItem['amount'] . ' currency: ' . $calculatedItem['currency']);
        }
        else {
          \Civi::log('stripeimport')
            ->info('StripeImport: ' . $this->getStripeSubscriptionID() . ' recurID: ' . $this->getContributionRecurID() . ': recur already updated: ' . $contributionRecur['id'] . '; amount: ' . $calculatedItem['amount'] . ' currency: ' . $calculatedItem['currency']);
        }
      }
    }
    else {
      // No existing contributionRecur - create one
      $contributionRecurParams = [
        'contact_id' => $this->getContactID(),
        'amount' => \CRM_Stripe_Api::getObjectParam('amount', $stripeSubscription),
        'currency' => \CRM_Stripe_Api::getObjectParam('currency', $stripeSubscription),
        'frequency_unit' => \CRM_Stripe_Api::getObjectParam('frequency_unit', $stripeSubscription),
        'frequency_interval' => \CRM_Stripe_Api::getObjectParam('frequency_interval', $stripeSubscription),
        'start_date' => \CRM_Stripe_Api::getObjectParam('plan_start', $stripeSubscription),
        'processor_id' => $this->getStripeSubscriptionID(),
        'cycle_day' => \CRM_Stripe_Api::getObjectParam('cycle_day', $stripeSubscription),
        'next_sched_contribution_date' => \CRM_Stripe_Api::getObjectParam('next_sched_contribution_date', $stripeSubscription),
        'auto_renew' => 1,
        'payment_processor_id' => $this->getPaymentProcessorID(),
        'payment_instrument_id' => $this->getPaymentInstrumentID(),
        'financial_type_id' => $this->getFinancialTypeID(),
        'is_email_receipt' => $this->getEmailReceipt(),
        'is_test' => $this->getPaymentProcessorObject()->getPaymentProcessor()['is_test'],
        'contribution_status_id:name' => \CRM_Stripe_Api::mapSubscriptionStatusToRecurStatus($stripeSubscription->status),
      ];
      // If we have a canceled_at date record it as end date
      $cancelDate = \CRM_Stripe_Api::getObjectParam('cancel_date', $stripeSubscription);
      if (!empty($cancelDate)) {
        $contributionRecurParams['end_date'] = $cancelDate;
      }

      $contributionRecur = ContributionRecur::create(FALSE)
        ->setValues($contributionRecurParams)
        ->execute()
        ->first();
      \Civi::log('stripeimport')->info('StripeImport: ' . $this->getStripeSubscriptionID() . ': Created Recur ID: ' . $contributionRecur['id'] . ': status: ' . $contributionRecurParams['contribution_status_id:name']);
      $this->setContributionRecurID($contributionRecur['id']);
    }

    // At this point we have both a ContactID and a ContributionRecurID

    // Link to membership record
    // By default we'll match the latest active membership, unless membership_id is passed in.
    if ($this->hasMembershipID()) {
      // Nothing to do. We have a MembershipID
    }
    elseif ($this->isMembershipAuto()) {
      // We need to "auto-detect" the MembershipID
      // Do that by firing an event 'civi.stripeimport.membershipauto'
      // A default listener in \Civi\StripeImport\MembershipSubscriber will select the newest Membership (by ID) for the Contact.
      // You can customise this by implementing your own listener for the event which returns a MembershipID.
      // If we don't return a MembershipID that's fine - we'll create Contribution/ContributionRecur with no Membership.
      $membershipAutoEvent = \Civi::dispatcher()->dispatch(
        'civi.stripeimport.membershipauto',
        new MembershipAutoEvent(
          $this->getContactID(),
          $this->getContributionRecurID(),
          $this->getPaymentProcessorObject()->getPaymentProcessor()['is_test'],
          $stripeSubscription
        )
      );
      if (!empty($membershipAutoEvent->membershipID)) {
        // Great, we've got a MembershipID to link to this subscription!
        $this->setMembershipID($membershipAutoEvent->membershipID);
      }
    }

    // Get the Stripe invoices for the subscription
    $invoiceParams = [
      'customer' => \CRM_Stripe_Api::getObjectParam('customer_id', $stripeSubscription),
      'subscription' => $this->getStripeSubscriptionID(),
      'limit' => 100,
    ];
    $stripeInvoices = $this->getPaymentProcessorObject()->stripeClient->invoices->all($invoiceParams);
    if ($stripeInvoices->count()) {
      // We have one or more invoices to import (as contributions)
      foreach ($stripeInvoices->data as $stripeInvoice) {
        $chargeID = \CRM_Stripe_Api::getObjectParam('charge_id', $stripeInvoice);
        if (empty($chargeID)) {
          continue;
        }

        $contributionID = NULL;

        // Check if we already have a Contribution for this Invoice
        // See if we have a matching payment for this trxn_id and get the linked contribution ID
        $contribution = civicrm_api3('Mjwpayment', 'get_contribution', [
          'trxn_id' => $chargeID,
        ]);
        if (!empty($contribution['id'])) {
          \Civi::log('stripeimport')->info('StripeImport: ' . $stripeInvoice->id . ': Found contributionID ' . $contribution['id'] . ' by matching Charge ID: ' . $chargeID);
        }

        if (empty($contribution['id'])) {
          // Try to find one based on date invoice was paid (any time during same day matches), amount, contact and financial type.
          $receiveDateOnly = date('Y-m-d', strtotime(\CRM_Stripe_Api::getObjectParam('receive_date', $stripeInvoice)));
          $invoiceDateOnly = date('Y-m-d', strtotime(\CRM_Stripe_Api::getObjectParam('invoice_date', $stripeInvoice)));
          $createdDateOnly = date('Y-m-d', strtotime(\CRM_Stripe_Api::getObjectParam('created_date', $stripeInvoice)));

          \Civi::log('stripeimport')->info('StripeImport: Looking for existing contribution with receive date: ' . $receiveDateOnly
            . ' or invoice date: ' . $invoiceDateOnly . ' or created date: ' . $createdDateOnly);
          // As we are matching an existing Contribution we're going to update/modify it.
          // So, Pending is ok. Also, the actual status is ok. But we don't want to eg. mark a Completed contribution as Failed.
          // Instead, we'd create a new Contribution with status=Failed.
          // It's quite likely, for example that there are Failed Invoices before a Completed one (eg. client was having trouble paying).
          $stripeInvoiceStatuses = ['Pending', \CRM_Stripe_Api::getObjectParam('status', $stripeInvoice)];
          $contribution = Contribution::get(FALSE)
            ->addSelect('financial_trxn.id', '*')
            ->addJoin('FinancialTrxn AS financial_trxn', 'LEFT', 'EntityFinancialTrxn')
            ->addWhere('financial_type_id', '=', $this->getFinancialTypeID())
            ->addWhere('total_amount', '=', \CRM_Stripe_Api::getObjectParam('amount', $stripeInvoice))
            ->addClause('OR',
              ['AND', [['receive_date', '>=', $receiveDateOnly . ' 00:00:00'], ['receive_date', '<=', $receiveDateOnly . ' 23:59:00']]],
              ['AND', [['receive_date', '>=', $invoiceDateOnly . ' 00:00:00'], ['receive_date', '<=', $invoiceDateOnly . ' 23:59:00']]],
              ['AND', [['receive_date', '>=', $createdDateOnly . ' 00:00:00'], ['receive_date', '<=', $createdDateOnly . ' 23:59:00']]])
            ->addOrderBy('receive_date', 'DESC')
            ->addWhere('contact_id', '=', $this->getContactID())
            ->addWhere('contribution_status_id:name', 'IN', $stripeInvoiceStatuses)
            ->execute()
            ->first();

          if (!empty($contribution['id'])) {
            // We've found a suitable contribution!
            \Civi::log('stripeimport')->info('StripeImport: ' . $stripeInvoice->id . ': Found contributionID ' . $contribution['id'] . ' by date/amount/financialType/status');
            if (!empty($contribution['financial_trxn.id'])) {
              // Update the existing financialtrxn with the charge/invoice ID.
              civicrm_api3('FinancialTrxn', 'create', [
                'id' => $contribution['financial_trxn.id'],
                'trxn_id' => \CRM_Stripe_Api::getObjectParam('charge_id', $stripeInvoice),
                'order_reference' => \CRM_Stripe_Api::getObjectParam('invoice_id', $stripeInvoice),
                'trxn_date' => \CRM_Stripe_Api::getObjectParam('receive_date', $stripeInvoice),
                'payment_processor_id' => $this->getPaymentProcessorID(),
              ]);
              // Update the contribution trxn_id with trxn_ids
              Contribution::update(FALSE)
                ->addValue('trxn_id', \CRM_Stripe_Api::getObjectParam('invoice_id', $stripeInvoice) . ',' . \CRM_Stripe_Api::getObjectParam('charge_id', $stripeInvoice))
                ->addValue('contribution_recur_id', $this->getContributionRecurID())
                ->addValue('receive_date', \CRM_Stripe_Api::getObjectParam('receive_date', $stripeInvoice))
                ->addWhere('id', '=', $contribution['id'])
                ->execute();
              if ($this->hasMembershipID()) {
                Membership::update(FALSE)
                  ->addValue('contribution_recur_id', $this->getContributionRecurID())
                  ->addWhere('id', '=', $this->getMembershipID())
                  ->execute();
              }
              else {
                \Civi::log('stripeimport')->info('StripeImport: ' . $stripeSubscription->id . ': No MembershipID for contactID: ' . $this->getContactID() . ' recurID: ' . $this->getContributionRecurID() . ' Recur status is: ' . \CRM_Stripe_Api::mapSubscriptionStatusToRecurStatus($stripeSubscription->status));
              }
            }
          }
        }

        // We have tried to find a Contribution
        if (!empty($contribution['id'])) {
          // and succeeded!
          $contributionID = $contribution['id'];
        }
        if (empty($contributionID)) {
          // We could not find a contribution for this invoice.
          // Use the Order API to create it and then we'll use the Webhook Processor to complete it.

          \Civi::log('stripeimport')->info('StripeImport: ' . $stripeInvoice->id . ': No existing contribution. Using Order API to create one');

          // Get the default price field ID
          if ($this->hasMembershipID()) {
            $priceFieldValues = \Civi\Api4\PriceFieldValue::getDefaultPriceFieldValueForMembershipMJW(FALSE)
              ->setMembershipID($this->getMembershipID())
              ->execute();
          }
          else {
            $priceFieldValues = \Civi\Api4\PriceFieldValue::getDefaultPriceFieldValueForContributionMJW(FALSE)->execute();
          }

          // Set up order API params.
          if (empty($this->getDescription())) {
            $this->setDescription(\CRM_Stripe_Api::getObjectParam('description', $stripeInvoice));
          }
          if (empty($this->getDescription()) && isset($stripeSubscription->items->first()->plan->product->name)) {
            $this->setDescription($stripeSubscription->items->first()->plan->product->name);
          }
          if (empty($this->getDescription())) {
            $this->setDescription(E::ts('StripeImport Subscription'));
          }
          \Civi::log('stripeimport')->info('StripeImport: ' . $stripeInvoice->id . ': Description: "' . $this->getDescription() . '"');

          $orderCreateParams = [
            'receive_date'           => \CRM_Stripe_Api::getObjectParam('receive_date', $stripeInvoice),
            // trxn_date is necessary for membership date calcs.
            'trxn_date'              => \CRM_Stripe_Api::getObjectParam('receive_date', $stripeInvoice),
            'trxn_id'                => \CRM_Stripe_Api::getObjectParam('invoice_id', $stripeInvoice),
            'total_amount'           => \CRM_Stripe_Api::getObjectParam('amount', $stripeInvoice),
            'contact_id'             => $this->getContactID(),
            'payment_instrument_id'  => $this->getPaymentInstrumentID(),
            'currency'               => \CRM_Stripe_Api::getObjectParam('currency', $stripeInvoice),
            'financial_type_id'      => $this->getFinancialTypeID(),
            'is_email_receipt'       => $this->getEmailReceipt(),
            'is_test'                => $this->getPaymentProcessorObject()->getPaymentProcessor()['is_test'],
            'contribution_recur_id'  => $this->getContributionRecurID(),
            'contribution_status_id' => 'Pending',
            'contribution_source'    => $this->getDescription(),
          ];
          $lineItemParams = [
            'contact_id' => $this->getContactID(),
          ];
          $lineItem = [
            'line_total' => \CRM_Stripe_Api::getObjectParam('amount', $stripeInvoice),
            'unit_price' => \CRM_Stripe_Api::getObjectParam('amount', $stripeInvoice),
            'price_field_id' => $priceFieldValues['price_field_id'],
            'price_field_value_id' => $priceFieldValues['price_field_value_id'],
            'financial_type_id' => $this->getFinancialTypeID(),
            'qty' => 1,
          ];
          if ($this->hasMembershipID()) {
            $lineItem['entity_table'] = 'civicrm_membership';
            $lineItemParams['id'] = $this->getMembershipID();
          }
          $orderCreateParams['line_items'] = [
            [
              'params' => $lineItemParams,
              'line_item' => [$lineItem]
            ]
          ];
          $contribution = civicrm_api3('Order', 'create', $orderCreateParams);
          $contributionID = $contribution['id'];
        }

        // At this point we definitely have a Contribution!

        // Now record the payment. We'll use the same logic as the webhook event processor
        $webhookEventProcessor = new \Civi\Stripe\Webhook\Events($this->getPaymentProcessorID());
        $webhookEventProcessor->setData(\Stripe\StripeObject::constructFrom(['object' => $stripeInvoice]));
        switch (\CRM_Stripe_Api::getObjectParam('status', $stripeInvoice)) {
          case 'Completed':
            if ($this->isMembershipDisableUpdates()) {
              $this->disableUpdateMembershipBasedOnCompletionOfContribution();
            }
            \Civi::log('stripeimport')->info('StripeImport: contributionID: ' . $contributionID . ' Triggering invoice.payment_succeeded webhook event');
            $webhookEventProcessor->setEventType('invoice.payment_succeeded');
            $webhookResult = $webhookEventProcessor->doInvoicePaid();
            \Civi::log('stripeimport')->info('StripeImport: contributionID: ' . $contributionID . ' invoice.payment_succeeded result: ' . ($webhookResult->ok ? 'SUCCESS' : 'ERROR') . ' Message: ' . $webhookResult->message);
            break;

          case 'Failed':
          case 'Cancelled':
            \Civi::log('stripeimport')->info('StripeImport: contributionID: ' . $contributionID . ' Triggering invoice.payment_failed webhook event');
            $webhookEventProcessor->setEventType('invoice.payment_failed');
            $webhookResult = $webhookEventProcessor->doInvoicePaymentFailed();
            \Civi::log('stripeimport')->info('StripeImport: contributionID: ' . $contributionID . ' invoice.payment_failed result: ' . ($webhookResult->ok ? 'SUCCESS' : 'ERROR') . ' Message: ' . $webhookResult->message);
            break;

          case 'Pending':
            // Nothing further to do (we already created the contribution and should not create a payment or failure yet)
        }
        $contributionIDs[] = $contributionID;
      }
    }
    else {
      // We have no invoices to import. Create a Template Contribution
      // This will be for one of the following reasons:
      //   - Stripe subscription is in a free "trial" period.
      //   - Stripe subscription has not yet reached the start date.
      // In this case we have to create a template contribution (see https://lab.civicrm.org/dev/financial/-/issues/6)
      //   because CiviCRM currently expects at least 1 contribution per ContributionRecur
      //   and it will allow a real contribution to be recorded when Stripe notifies us by webhook.

      \Civi::log('stripeimport')->info('StripeImport: ' . $stripeSubscription->id . ': No invoices for subscription. Creating template contribution');

      $contribution = Contribution::create(FALSE)
        ->addValue('contribution_recur_id', $contributionRecur['id'])
        ->addValue('contact_id', $contributionRecur['contact_id'])
        ->addValue('financial_type_id', $contributionRecur['financial_type_id'])
        ->addValue('payment_instrument_id', $contributionRecur['payment_instrument_id'])
        ->addValue('source', $this->getDescription())
        ->addValue('total_amount', $contributionRecur['amount'])
        ->addValue('currency', $contributionRecur['currency'])
        ->addValue('is_test', $this->getPaymentProcessorObject()->getPaymentProcessor()['is_test'])
        ->addValue('is_template', TRUE)
        ->addValue('contribution_status_id:name', 'Template')
        ->execute()
        ->first();
      $contributionID = $contribution['id'];
      if ($this->hasMembershipID() && !empty($contributionID)) {
        try {
          Membership::linkToRecurMJW(FALSE)
            ->addValue('id', $this->getMembershipID())
            ->addValue('contribution_id', $contributionID)
            ->execute();
        }
        catch(\CRM_Core_Exception $e) {
          \Civi::log('stripeimport')->warning('StripeImport: ' . $this->getStripeSubscriptionID() . ' ' . $e->getMessage());
        }
      }

      $contributionIDs[] = $contributionID;
    }
    $results = [
      'subscription' => $this->getStripeSubscriptionID(),
      'customer' => \CRM_Stripe_Api::getObjectParam('customer_id', $stripeSubscription),
      'contact_id' => $this->getContactID(),
      'recur_id' => $this->getContributionRecurID(),
      'contribution_ids' => $contributionIDs ?? [],
      'membership_id' => $this->hasMembershipID() ? $this->getMembershipID() : NULL,
    ];
    \Civi::log('stripeimport')->info('StripeImport: ' . $this->getStripeSubscriptionID() . ': END ' . print_r($results, TRUE));
    return $results;
  }

  /**
   * @param array $stripeAPIFilters
   *   eg. ['limit' => 100, 'created' => ['>' => 20231202000000], 'starting_after' => 'sub_xxx']
   *   'created' accepts anything that strtotime accepts - eg. '-1 week'
   *
   * @return array
   * @throws \Exception
   */
  public function importMultiple(array $stripeAPIFilters = []): array {
    // Get the subscriptions from Stripe
    if (empty($stripeAPIFilters['limit'])) {
      $stripeAPIFilters['limit'] = 100;
    }
    if (empty($stripeAPIFilters['status'])) {
      $stripeAPIFilters['status'] = 'all';
    }

    foreach ($stripeAPIFilters as $filterKey => $filterValue) {
      switch ($filterKey) {
        case 'starting_after':
          if (empty($filterValue)) {
            unset($stripeAPIFilters['starting_after']);
          }
          break;

        case 'created_after':
          if (is_array($filterValue)) {
            throw new \CRM_Core_Exception('created_after cannot be an array');
          }
          // Don't add a blank filter value
          if (empty($filterValue)) {
            unset($stripeAPIFilters['created_after']);
            break;
          }

          $stripeAPIFilters['created']['gt'] = strtotime($filterValue);
          unset($stripeAPIFilters['created_after']);
          // $stripeAPIFilters['created']['gte'];
          break;

        case 'created_before':
          if (is_array($filterValue)) {
            throw new \CRM_Core_Exception('created_after cannot be an array');
          }
          // Don't add a blank filter value
          if (empty($filterValue)) {
            unset($stripeAPIFilters['created_before']);
            break;
          }

          $stripeAPIFilters['created']['lt'] = strtotime($filterValue);
          // $stripeAPIFilters['created']['lte'];
          unset($stripeAPIFilters['created_before']);
          break;
      }

    }

    $stripeSubscriptions = $this->getPaymentProcessorObject()->stripeClient->subscriptions->all($stripeAPIFilters);

    $stripeSubscriptionIDs = array_map(
      function ($subscription) { return $subscription->id; },
      $stripeSubscriptions->data
    );

    // Prepare to collect the results
    $results = [
      'imported' => [],
      'errors' => [],
      'continue_after' => NULL
    ];

    // Exit if there aren't records to process
    if (!count($stripeSubscriptionIDs)) {
      return $results;
    }

    foreach ($stripeSubscriptionIDs as $subscriptionID) {
      $this->resetOne();
      $this->setStripeSubscriptionID($subscriptionID);

      try {
        $resultImportOne = $this->importOne();
        $results['imported'][$subscriptionID] = $resultImportOne;
      }
      catch (\Exception $e) {
        \Civi::log('stripeimport')->error('StripeImport: ' . $subscriptionID . ' Import Failed: ' . $e->getMessage());
        $results['errors'][$subscriptionID] = $e->getMessage();
      }
    }
    $results['continue_after'] = $subscriptionID;
    $results['count'] = [
      'imported' => count($results['imported']),
      'errors' => count($results['errors']),
    ];
    return $results;
  }

}
